/*
 Simple FAT12 implementation

 This code is released into the public domain
 */

#include <stdio.h>
#include <stdlib.h>

#include <fcntl.h>
#ifdef _MSC_VER
#include <wchar.h>
#endif

#include "error.h"
#include "fat12.h"

#ifndef O_BINARY
#define O_BINARY 0
#endif

static int fat_read(fatso * ft, byte * buf, int offset, int count)
{
    off_t pos, newpos;
    int   nbytes, expected;

    pos = offset * ft->BytsPerSec;
    newpos = lseek(ft->fd,pos, SEEK_SET);
    if (newpos == -1)
    {
        error(ERROR_STDLIB,"Unable to seek to %lx.",pos);
        return -1;
    }
    expected = count*ft->BytsPerSec;
    nbytes = read(ft->fd, buf, expected);
    if (nbytes != expected)
    {
        if (nbytes < 0) {
            error(ERROR_STDLIB, "reading %d bytes from offset %d.",
                pos, count*ft->BytsPerSec);
        }
        else
        {
            error(0,"Only read %d bytes out of %d at offset %ld!\n",
                nbytes, count*ft->BytsPerSec, newpos);

        }
        return -1;
    }
    return 0;
}
   
int fat_birth(fatso * ft, char * fname)
{
    byte    BPB[512];
    int     TotSec32, TotSec16, TotSec;

    memset(ft, 0, sizeof(struct fatso));
    ft->magic = fats_magic;

    ft->fd = open(fname, O_RDONLY|O_BINARY);
    if (ft->fd < 0) {
        error(ERROR_STDLIB,"Unable to open file \"%s\".", fname);
        return -1;
    }
    ft->BytsPerSec = 512;
    if (fat_read(ft,BPB,0,1) != 0) {
        fat_death(ft);
        return -1;
    }
    ft->BytsPerSec = READ_LE_WORD(&BPB[11]);
    ft->SecPerClus = BPB[13];
    ft->RsvdSecCnt = READ_LE_WORD(&BPB[14]);
    ft->NumFATs = BPB[16];
    ft->RootEntCnt = READ_LE_WORD(&BPB[17]);
    TotSec16 = READ_LE_WORD(&BPB[19]);
    ft->FATSz16 = READ_LE_WORD(&BPB[22]);
    TotSec32 = READ_LE_DWORD(&BPB[32]);
#if 0
    printf("Bytes Per Sector...............  %d \n", ft->BytsPerSec);
    printf("Sectors Per Cluster............  %d \n", ft->SecPerClus);
    printf("Reserved Sector Count..........  %d \n", ft->RsvdSecCnt);
    printf("Number of FATs.................  %d \n", ft->NumFATs);
    printf("Root Entry Count...............  %d \n", ft->RootEntCnt);
    printf("Total Size (sectors) 16 bit....  %d \n", TotSec16);
    printf("FAT Size (sectors).............  %d \n", ft->FATSz16);
    printf("Total Size (sectors) 32 bit....  %lu (%lx) \n", TotSec32,TotSec32);
    printf("--------------------------------------\n");
#endif
    if ((!ft->FATSz16) ||  (!ft->BytsPerSec))
    {
        error(0,"FAT information doesn't appear valid.\n");
        fat_death(ft);
        return -1;
    }
    ft->RootDirSectors = ((ft->RootEntCnt * 32) + (ft->BytsPerSec - 1))/
        ft->BytsPerSec;
    ft->FirstDataSector = ft->RsvdSecCnt + (ft->NumFATs * ft->FATSz16) + 
        ft->RootDirSectors;
#if 0
    printf("Root Directory Sectors.........  %d \n", ft->RootDirSectors);
    printf("First Data Sector..............  %d \n", ft->FirstDataSector);
#endif

    if (TotSec16 != 0)  TotSec = TotSec16;
    else TotSec = TotSec32;
    if (!TotSec) {
        error(0, "FAT image doesn't appear to have any sectors!");
        fat_death(ft);
        return -1;
    }
    ft->CountOfClusters = (TotSec - ft->FirstDataSector)/ft->SecPerClus;
#if 0
    printf("Count Of Clusters..............  %d \n", ft->CountOfClusters);
#endif
    if (ft->CountOfClusters > 4085)
    {
        error(0, "Can't handle a FAT image with %d clusters!",
            ft->CountOfClusters);
        fat_death(ft);
        return -1;
    }
    ft->fat = malloc(ft->FATSz16 * ft->BytsPerSec);
    if (!ft->fat) {
        error(0,"Not enough memory for FAT buffer!");
        fat_death(ft);
        return -1;
    }
    if (fat_read(ft,ft->fat,ft->RsvdSecCnt,ft->FATSz16) != 0) { 
        fat_death(ft); return(-1);
    }
    ft->magic = fats_magic;
    return 0;
}


void fat_death(fatso * ft)
{
    if (!ft || (ft->magic != fats_magic)) return;

    if (ft->fat) free(ft->fat);
    if (ft->fd) close(ft->fd);
    memset(ft, 0, sizeof(struct fatso));
    return;
}


int fat_next_cluster(fatso * ft, int cluster)
{
    int offset;
    int next;

    if (cluster > ft->CountOfClusters) return -1;
    next = READ_LE_WORD(&ft->fat[ (cluster * 3)/2 ] );
    if (cluster & 1) next >>= 4;
    else next &= 0xfff;
    if ((next > 0xff8) || ( next < 2)) return -1;
    return next;
}

fatdir * fat_directory(fatso * ft, fatentry * fe)
{
    fatdir * fd;
    byte   * p;
    int    BytsPerClus, size;

    BytsPerClus = ft->BytsPerSec*ft->SecPerClus;

    if (!fe) {
        size = ft->RootDirSectors*ft->BytsPerSec;
        fd = malloc(sizeof(struct fatdir) + size);
        if (!fd) {
            error(0,"Unable to malloc directory buffer.");
            return NULL;
        }
        fd->size = size;
        p = (byte *)fd + sizeof(struct fatdir);
        if (fat_read(ft,p,ft->FirstDataSector-ft->RootDirSectors, 
        ft->RootDirSectors) != 0) {
            free(fd);
            return NULL;
        }
    }

    else {
        int cluster = fe->cluster;
        int count;

        size = 0;
        count = 0;
        while (cluster != -1) {
            size += BytsPerClus;    
            cluster = fat_next_cluster(ft,cluster);
            if (count > ft->CountOfClusters) {
                error(0,"Error reading FAT image - Looped FAT!");
                return NULL;
            }
        }
        cluster = fe->cluster;
        fd = malloc(sizeof(struct fatdir) + size);
        p = (byte * )fd + sizeof(struct fatdir); 
        while (cluster != -1) {
            if (fat_read(ft,p,ft->FirstDataSector + (cluster-2)*ft->SecPerClus,
            ft->SecPerClus) != 0) {
                free(fd);
                return NULL;
            }
            p += BytsPerClus;
            cluster = fat_next_cluster(ft,cluster);
        } fd->size = size;
    }
    fd->magic = fdir_magic;
    return fd;
}

int fat_entry(fatdir * fd, int index, fatentry * fe)
{
    byte * p; 
    int i;

    if (!fe || !fd || (fd->magic != fdir_magic)) {
        return -1;
    }
    if (((index*32)+31) > fd->size) return 1;
    p = (byte *)fd + sizeof(struct fatdir);
    p += index * 32;
    memset(fe, 0, sizeof(fe));
    if (p[0] == '\0') return 1;
    memcpy(fe->name, p, 12);

    /* Convert spaces to 0's to get better strlen results */
    fe->name[11] = '\0';
    for (i = 10; i > 0; i--) {
        if (fe->name[i] == ' ') fe->name[i] = '\0'; 
        else break;
    }
    fe->attributes = *(p+11);
    fe->cluster = READ_LE_WORD(p+26);
    fe->size  = READ_LE_DWORD(p+28);
    return 0;
}

byte * fat_file(fatso * ft, fatentry * fe)
{
    byte * buf, * p;
    int cluster, clusters;
    int BytsPerClus, count;

    BytsPerClus = ft->BytsPerSec*ft->SecPerClus;

    if (!fe->size) {
        error(0,"Attempt to read \"%s\" from FAT image - File is empty.",
            fe->name);
        return NULL;
    }
    clusters = (fe->size + (BytsPerClus-1))/BytsPerClus;
    buf = malloc((clusters * BytsPerClus) + 4);
    if (!buf) {
        error(0,"Unable to allocate memory for file from FAT image.");
        return NULL;
    }
    p = buf + sizeof(dword);
    *(dword *)buf = fe->size;
    count = 0;
    cluster = fe->cluster;
    while (cluster != -1) {
        if (fat_read(ft,p,ft->FirstDataSector + (cluster-2)*ft->SecPerClus,
        ft->SecPerClus) != 0) {
            free(buf);
            return NULL;
        }
        p += BytsPerClus;         
 
        cluster = fat_next_cluster(ft,cluster);
        if (count > ft->CountOfClusters) {
            error(0,"Error reading FAT image - Looped FAT!");
            free(buf);
            return NULL;
        }
    }
    return buf;
}

#if 0
int fat_check_filename(fatentry * fe)
{
    switch (fe->name[0])
    {
    case 0xE5: return 0;
    case 0:    return 0;
    default:    ;
    }
    return 1; 
}

int rec_list( fatso * fat, int depth, fatdir * dir)
{
    int index, r;
    fatentry    fe;
    int i;

    printf("Starting with %lx\n", dir);
    if (!dir) return;

    index = 0;
    do {
        r = fat_entry(dir, index++, &fe);
        if (!r && fat_check_filename(&fe)) {
            for (i = 0; i < depth; i++) { printf("   "); }
            printf("%.11s (%0x) %d bytes, at %x.\n", fe.name, fe.attributes,
                fe.size, fe.cluster);
            if ((strcmp(fe.name,".") != 0) && (strcmp(fe.name,"..") != 0) &&
            (fe.attributes & 0x10)) {
                fatdir * newd;

                rec_list( fat, depth + 1, fat_directory(fat, &fe));
                printf("Here...\n");
            }
            if (((fe.attributes & 0x10) == 0) && (fe.size)) {
                byte * p;

                p = fat_file(fat,&fe);
                if (p)  {
                    hexdump(p, fe.size);
                    free(p);
                }
            }
        }
    } while (r == 0);
    printf("Freeing %lx\n", dir);
    free(dir);
    return 0;
}


int main(int argc, char ** argv)
{
    fatso   fat;
    int     r;

    if (argc != 2) {
        fprintf(stderr,"Usage: FAT image filename.\n");
        exit(-1);
    }
    r = fat_birth(&fat, argv[1]);
    printf("Birth returned %d\n", r);
    if (r) exit(-1);
    rec_list( &fat, 0, fat_directory(&fat, NULL));
    printf("..done\n");
    fat_death(&fat);
    return 0;
}
#endif
